@@ -40,8 +40,10 @@ export objdir=${srcdir}/.build
 export instdir=${srcdir}/.inst
 export srcinstdir=${srcdir}/.sinst
 export checkfile=${topdir}/${FULLPKG}.check
-host=i686-pc-cygwin
-target=i686-pc-cygwin
+#host=i686-pc-cygwin
+#host=x86_64-pc-cygwin
+#target=i686-pc-cygwin
+#target=x86_64-pc-cygwin
 prefix=/usr
 pnoslash=`echo $prefix | sed -e 's#^/##'`
 sysconfdir=/etc
@@ -61,16 +63,39 @@ mkdirs() {
 prep() {
   (cd ${topdir} && \
   tar xvjf ${src_orig_pkg} && \
-  tar xvjf ${src_orig_doc_pkg} && \
-  cd ${topdir} && \
+  if [ -f ${src_orig_doc_pkg} ] ; then \
+    tar xvjf ${src_orig_doc_pkg} ; \
+  fi && \
+  cd ${srcdir} && \
   if [ -f ${src_patch} ] ; then \
-    patch -p0 < ${src_patch} ; \
+    patch -p1 < ${src_patch} ; \
+  fi && \
+  cd ${topdir} && \
+  mkdirs ) 
+}
+devprep() {
+  (cd ${topdir} && \
+  tar xvjf ${src_orig_pkg} && \
+  if [ -f ${src_orig_doc_pkg} ] ; then \
+    tar xvjf ${src_orig_doc_pkg} ; \
+  fi && \
+  mv ${FULLPKG} ${PKG}-${VER} && \
+  cd ${srcdir} && \
+  if [ -f ${src_patch} ] ; then \
+    patch -p1 < ${src_patch} ; \
   fi && \
   mkdirs ) 
 }
+doautoconf() {
+  (cd ${srcdir} && \
+   rm -rf autom4te.cache && \
+   echo "Running autoconf" && autoconf && \
+   echo "Running autoheader" && autoheader && \
+   echo > stamp-h.in)
+}
 conf() {
   (cd ${objdir} && \
-  ${srcdir}/configure --host=${host} --target=${target} \
+  ${srcdir}/configure \
   --srcdir=${srcdir} \
   --prefix=${prefix} \
   --sysconfdir=${sysconfdir} \
@@ -78,6 +103,8 @@ conf() {
   --mandir=${mandir} \
   --localstatedir=/var \
   --enable-symlinks \
+  --enable-pcre \
+  --enable-multibyte \
   --with-curses-terminfo \
   --with-install-prefix=${instdir} \
   ${ZSHCONFDEBUG} )
@@ -112,7 +139,8 @@ install() {
     rm ${instdir}${prefix}/share/info/dir ; \
   fi && \
   for dirn in ${prefix}/share/doc/Cygwin ${prefix}/share/doc/${PKG}-${VER} \
-              ${sysconfdir}/postinstall ${sysconfdir}/profile.d ; do \
+              ${sysconfdir}/postinstall ${sysconfdir}/preremove \
+              ${sysconfdir}/profile.d ; do \
     if [ ! -d ${instdir}${dirn} ]; then \
       mkdir -p ${instdir}${dirn} ; \
     fi ; \
@@ -130,30 +158,50 @@ install() {
   fi && \
   echo "Copying example Functions and StartupFiles" && \
   cp -ar ${srcdir}/Functions ${instdir}${prefix}/share/doc/${PKG}-${VER} && \
-  cp -ar ${srcdir}/StartupFiles ${instdir}${prefix}/share/doc/${PKG}-${VER} && \
+  mkdir -p ${instdir}${prefix}/share/doc/${PKG}-${VER}/StartupFiles/etc && \
+  cp -ar ${srcdir}/StartupFiles/* ${instdir}${prefix}/share/doc/${PKG}-${VER}/StartupFiles/etc && \
   templist=""; \
-  for f in ${srcdir}/CYGWIN-PATCHES/zprofile ${srcdir}/CYGWIN-PATCHES/zlogin \
-           ${srcdir}/CYGWIN-PATCHES/zlogout ${srcdir}/CYGWIN-PATCHES/zshenv \
-           ${srcdir}/CYGWIN-PATCHES/zshrc \
-           ${srcdir}/CYGWIN-PATCHES/.zprofile ${srcdir}/CYGWIN-PATCHES/.zlogin \
-           ${srcdir}/CYGWIN-PATCHES/.zlogout ${srcdir}/CYGWIN-PATCHES/.zshenv \
-           ${srcdir}/CYGWIN-PATCHES/.zshrc ; do \
+  for f in ${srcdir}/CYGWIN-PATCHES/StartupFiles/etc/* ; do \
     if [ -f $f ] ; then \
       templist="$templist $f"; \
     fi ; \
   done && \
   if [ ! "x$templist" = "x" ]; then \
     /usr/bin/install -c -m 644 $templist \
-         ${instdir}${prefix}/share/doc/${PKG}-${VER}/StartupFiles ; \
+         ${instdir}${prefix}/share/doc/${PKG}-${VER}/StartupFiles/etc ; \
   fi && \
-  /usr/bin/install -c -m 755 ${srcdir}/CYGWIN-PATCHES/mkzsh \
-    ${instdir}${prefix}/bin && \
-  /usr/bin/install -c -m 755 ${srcdir}/CYGWIN-PATCHES/profile.d/* \
-    ${instdir}${sysconfdir}/profile.d && \
-  echo "Generating postinstall/${PKG}.sh" && \
-  sed "s/%VERSION%/${VER}/" < ${srcdir}/CYGWIN-PATCHES/postinstall.sh > /tmp/.inst.postinstall && \
-  /usr/bin/install -c -m 755 /tmp/.inst.postinstall ${instdir}${sysconfdir}/postinstall/${PKG}.sh && \
-  rm -f /tmp/.inst.postinstall && \
+  templist=""; \
+  for f in ${srcdir}/CYGWIN-PATCHES/bin/* ; do \
+    if [ -f $f ] ; then \
+      templist="$templist $f"; \
+    fi ; \
+  done && \
+  if [ ! "x$templist" = "x" ]; then \
+    /usr/bin/install -c -m 755 $templist ${instdir}${prefix}/bin ; \
+  fi && \
+  templist=""; \
+  for f in ${srcdir}/CYGWIN-PATCHES/profile.d/* ; do \
+    if [ -f $f ] ; then \
+      templist="$templist $f"; \
+    fi ; \
+  done && \
+  if [ ! "x$templist" = "x" ]; then \
+    /usr/bin/install -c -m 755 $templist ${instdir}${sysconfdir}/profile.d ; \
+  fi && \
+  echo "Generating postinstall files" && \
+  for f in ${srcdir}/CYGWIN-PATCHES/postinstall/* ; do \
+    templist=`basename ${f}`; \
+    sed "s/%VERSION%/${VER}/" < ${f} > /tmp/.inst.postinstall && \
+    /usr/bin/install -c -m 755 /tmp/.inst.postinstall ${instdir}${sysconfdir}/postinstall/${templist} && \
+    rm -f /tmp/.inst.postinstall ; \
+  done && \
+  echo "Generating preremove files" && \
+  for f in ${srcdir}/CYGWIN-PATCHES/preremove/* ; do \
+    templist=`basename ${f}`; \
+    sed "s/%VERSION%/${VER}/" < ${f} > /tmp/.inst.preremove && \
+    /usr/bin/install -c -m 755 /tmp/.inst.preremove ${instdir}${sysconfdir}/preremove/${templist} && \
+    rm -f /tmp/.inst.preremove ; \
+  done && \
   echo "Generating ${PKG}-${VER}.README" && \
   /usr/bin/install -c -m 644 ${srcdir}/CYGWIN-PATCHES/${PKG}.README \
     ${instdir}${prefix}/share/doc/Cygwin/${PKG}-${VER}.README && \
@@ -166,7 +214,9 @@ install() {
 mkpatch() {
   (cd ${srcdir} && \
   tar xvjf ${src_orig_pkg} && \
-  tar xvjf ${src_orig_doc_pkg} && \
+  if [ -f ${src_orig_doc_pkg} ] ; then \
+    tar xvjf ${src_orig_doc_pkg} ; \
+  fi && \
   mv ${PKG}-${VER} ../${PKG}-${VER}-orig && \
   cd ${topdir} && \
   diff -urN -x '.build' -x '.inst' -x '.sinst' \
@@ -174,6 +224,19 @@ mkpatch() {
     ${srcinstdir}/${src_patch_name} ; \
   rm -rf ${PKG}-${VER}-orig )
 }
+devmkpatch() {
+  (cd ${srcdir} && \
+  tar xvjf ${src_orig_pkg} && \
+  if [ -f ${src_orig_doc_pkg} ] ; then \
+    tar xvjf ${src_orig_doc_pkg} ; \
+  fi && \
+  mv ${FULLPKG} ../${PKG}-${VER}-orig && \
+  cd ${topdir} && \
+  diff -urN -x '.build' -x '.inst' -x '.sinst' \
+    ${PKG}-${VER}-orig ${PKG}-${VER} > \
+    ${srcinstdir}/${src_patch_name} ; \
+  rm -rf ${PKG}-${VER}-orig )
+}
 strip() {
   (cd ${instdir} && \
   find . -name "*.exe" | xargs strip > /dev/null 2>&1 
@@ -186,7 +249,19 @@ pkg() {
 spkg() {
   (mkpatch && \
   cp ${src_orig_pkg} ${srcinstdir}/${src_orig_pkg_name} && \
-  cp ${src_orig_doc_pkg} ${srcinstdir}/${src_orig_doc_pkg_name} && \
+  if [ -f ${src_orig_doc_pkg} ] ; then \
+    cp ${src_orig_doc_pkg} ${srcinstdir}/${src_orig_doc_pkg_name} ; \
+  fi && \
+  cp $0 ${srcinstdir}/`basename $0` && \
+  cd ${srcinstdir} && \
+  tar cvjf ${src_pkg} * )
+}
+devspkg() {
+  (devmkpatch && \
+  cp ${src_orig_pkg} ${srcinstdir}/${src_orig_pkg_name} && \
+  if [ -f ${src_orig_doc_pkg} ] ; then \
+    cp ${src_orig_doc_pkg} ${srcinstdir}/${src_orig_doc_pkg_name} ; \
+  fi && \
   cp $0 ${srcinstdir}/`basename $0` && \
   cd ${srcinstdir} && \
   tar cvjf ${src_pkg} * )
@@ -196,7 +271,9 @@ finish() {
 }
 case $1 in
   prep)	prep ; STATUS=$? ;;
+  devprep)	devprep ; STATUS=$? ;;
   mkdirs)	mkdirs ; STATUS=$? ;;
+  autoconf)	doautoconf ; STATUS=$? ;;
   conf)	conf ; STATUS=$? ;;
   build)	build ; STATUS=$? ;;
   check)	check ; STATUS=$? ;;
@@ -204,10 +281,12 @@ case $1 in
   install)	install ; STATUS=$? ;;
   strip)	strip ; STATUS=$? ;;
   mkpatch) mkpatch ; STATUS=$? ;;
+  devmkpatch) devmkpatch ; STATUS=$? ;;
   package)	pkg ; STATUS=$? ;;
   pkg)	pkg ; STATUS=$? ;;
   src-package)	spkg ; STATUS=$? ;;
   spkg)	spkg ; STATUS=$? ;;
+  devspkg)	devspkg ; STATUS=$? ;;
   finish) finish ; STATUS=$? ;;
   all) prep && conf && build && install && \
      pkg && spkg && finish ; STATUS=$? ;;
